# SPDX-FileCopyrightText: © 2026 Tenstorrent Inc.
# SPDX-License-Identifier: Apache-2.0

import pytest

from typing import List, Dict, Tuple, Union

from pyluwen import pci_scan, PciChip
from tt_umd import PCIDevice, TopologyDiscovery, TTDevice
from tt_smi.tt_smi_backend import TTSMIBackend
from tt_tools_common.utils_common.tools_utils import detect_chips_with_callback

# Luwen fixtures


@pytest.fixture(scope="session")
def luwen_reset_test_config() -> Tuple[List[int], bool]:
    """Return a tuple of a list of PCI indices using Luwen backend, and False, which represents the use_umd param."""
    return (pci_scan(), False)


@pytest.fixture(scope="session")
def luwen_devices() -> Dict[int, PciChip]:
    """Return a list of Tenstorrent PciChips."""
    return dict(enumerate(detect_chips_with_callback()))


@pytest.fixture(scope="session")
def luwen_backend(luwen_devices) -> TTSMIBackend:
    """Return a TTSMIBackend instance created from devices."""
    return TTSMIBackend(luwen_devices)


# UMD fixtures


@pytest.fixture(scope="session")
def umd_reset_test_config() -> Tuple[List[int], bool]:
    """Return a tuple of a list of PCI indices using UMD backend, and True, which represents the use_umd param."""
    return (list(PCIDevice.enumerate_devices_info()), True)


@pytest.fixture(scope="session")
def umd_devices() -> dict[int, TTDevice]:
    """Return a dict of Tenstorrent TTDevices using UMD backend."""
    _, devices = TopologyDiscovery.discover()
    return devices


@pytest.fixture(scope="session")
def umd_backend() -> TTSMIBackend:
    """Return a TTSMIBackend instance using UMD backend."""
    cluster_descriptor, devices = TopologyDiscovery.discover()
    return TTSMIBackend(devices=devices, umd_cluster_descriptor=cluster_descriptor)


# Parametrized fixtures to run tests with both backends


@pytest.fixture(scope="session", params=["luwen", "umd"])
def reset_test_config(
    request, luwen_reset_test_config, umd_reset_test_config
) -> Tuple[List[int], bool]:
    if request.param == "luwen":
        return luwen_reset_test_config
    return umd_reset_test_config


@pytest.fixture(scope="session", params=["luwen", "umd"])
def devices(request, luwen_devices, umd_devices) -> List[Union[PciChip, TTDevice]]:
    if request.param == "luwen":
        return luwen_devices
    return umd_devices


@pytest.fixture(scope="session", params=["luwen", "umd"])
def backend(request, luwen_backend, umd_backend) -> TTSMIBackend:
    if request.param == "luwen":
        return luwen_backend
    return umd_backend



# SPDX-FileCopyrightText: © 2026 Tenstorrent Inc.
# SPDX-License-Identifier: Apache-2.0

"""
Reset-related functions for tt-smi: PCI board reset, galaxy 6U tray reset, and helpers.
"""

import os
import sys
import time
from typing import List

from tt_tools_common.ui_common.themes import CMD_LINE_COLOR
from tt_tools_common.reset_common.wh_reset import WHChipReset
from tt_tools_common.reset_common.bh_reset import BHChipReset
from pyluwen import (
    PciChip,
    run_wh_ubb_ipmi_reset,
    run_ubb_wait_for_driver_load,
)
from tt_umd import (
    WarmReset,
    PCIDevice,
    TopologyDiscovery,
    TopologyDiscoveryOptions,
)
from tt_tools_common.utils_common.tools_utils import (
    detect_chips_with_callback,
)


def timed_wait(seconds):
    print("\033[93mWaiting for {} seconds: 0\033[0m".format(seconds), end='')
    sys.stdout.flush()

    for i in range(1, seconds + 1):
        time.sleep(1)
        print("\r\033[93mWaiting for {} seconds: {}\033[0m".format(seconds, i), end='')
        sys.stdout.flush()
    print()

# Keep this function for now, but it's not used anywhere in the codebase. 
# It is the check that is used in Funtest for Galaxy systems and we might need to reference it later.
def check_wh_galaxy_eth_link_status(devices):
    """
    Check the WH Galaxy Ethernet link status.
    Returns True if the link is up, False otherwise.
    """
    noc_id = 0
    DEBUG_BUF_ADDR = 0x12c0  # For eth fw 5.0.0 and above
    eth_locations_noc_0 = [
        (9, 0), (1, 0), (8, 0), (2, 0), (7, 0), (3, 0), (6, 0), (4, 0),
        (9, 6), (1, 6), (8, 6), (2, 6), (7, 6), (3, 6), (6, 6), (4, 6),
    ]
    LINK_INACTIVE_FAIL_DUMMY_PACKET = 10
    if len(devices) != 32:
        print(
            CMD_LINE_COLOR.RED,
            f"Error: Expected 32 devices for WH Galaxy Ethernet link status check, seeing {len(devices)}, please try reset again or cold boot the system.",
            CMD_LINE_COLOR.ENDC,
        )
        sys.exit(1)

    link_errors = {}
    for i, device in enumerate(devices):
        for eth in range(16):
            eth_x, eth_y = eth_locations_noc_0[eth]
            link_error = device.noc_read32(noc_id, eth_x, eth_y, DEBUG_BUF_ADDR + 0x4 * 96)
            if link_error == LINK_INACTIVE_FAIL_DUMMY_PACKET:
                link_errors[i] = eth

    if link_errors:
        for board_idx, eth in link_errors.items():
            print(
                CMD_LINE_COLOR.RED,
                f"Board {board_idx} has link error on eth port {eth}",
                CMD_LINE_COLOR.ENDC,
            )
        raise Exception("WH Galaxy Ethernet link errors detected!")

def umd_ubb_wait_for_driver_load():
    """
    Wait for the driver to reload for UMD, try 100 times.
    Similar to luwen's ubb_wait_for_driver_load but uses PCIDevice.enumerate_devices.
    """
    attempts = 0
    expected_chip_count = 32

    while attempts < 100:
        device_count = 0
        try:
            devices = PCIDevice.enumerate_devices()
            device_count = len(devices)
            if device_count == expected_chip_count:
                print(f"Driver loaded with {device_count} devices")
                return
        except Exception:
            pass

        print(f"Waiting for driver load ... {attempts} seconds (found {device_count} devices)")
        time.sleep(1)
        attempts += 1

    raise Exception(
        f"Driver not loaded with {expected_chip_count} devices after 100 seconds... giving up"
    )


def pci_board_reset(
    list_of_boards: List[int],
    reinit: bool = False,
    print_status: bool = True,
    use_umd: bool = False,
    wait_for_eth: bool = True,
):
    """Given a list of PCI index's init the PCI chip and call reset on it"""
    reset_wh_pci_idx = []
    reset_bh_pci_idx = []
    board_types = set()
    if use_umd:
        chips = PCIDevice.enumerate_devices_info()
    for pci_idx in list_of_boards:
        try:
            if not use_umd:
                chip = PciChip(pci_interface=pci_idx)
        except Exception as e:
            print(e)
            print(
                CMD_LINE_COLOR.RED,
                f"Error accessing board at PCI index {pci_idx}! Use -ls to see all devices available to reset",
                CMD_LINE_COLOR.ENDC,
            )
            continue

        if use_umd:
            board_types.add(chips[pci_idx].subsystem_id)
        else:
            if chip.as_wh():
                reset_wh_pci_idx.append(pci_idx)
                board_types.add(chip.as_wh().pci_board_type())
            elif chip.as_bh():
                reset_bh_pci_idx.append(pci_idx)
                board_types.add(chip.as_bh().pci_board_type())
            else:
                print(
                    CMD_LINE_COLOR.RED,
                    "Unknown chip type detected. Exiting...",
                    CMD_LINE_COLOR.ENDC,
                )
                del chip
                sys.exit(1)
            del chip

    is_galaxy = board_types <= {0x35, 0x47}
    if is_galaxy:
        print(
            CMD_LINE_COLOR.YELLOW,
            "CPLD FW v1.16 or higher is required to use tt-smi -r on Galaxy systems.",
            "If tt-smi -r fails, please continue to use tt-smi -glx_reset instead and contact your system administrator to request a CPLD update.",
            CMD_LINE_COLOR.ENDC,
        )

    secondary_bus_reset = not is_galaxy

    if use_umd:
        WarmReset.warm_reset(list_of_boards, secondary_bus_reset=secondary_bus_reset)
    else:
        if reset_wh_pci_idx:
            WHChipReset().full_lds_reset(
                pci_interfaces=reset_wh_pci_idx, secondary_bus_reset=secondary_bus_reset
            )
        if reset_bh_pci_idx:
            BHChipReset().full_lds_reset(
                pci_interfaces=reset_bh_pci_idx, secondary_bus_reset=secondary_bus_reset
            )

    if reinit:
        os.environ["RUST_BACKTRACE"] = "full"
        print(
            CMD_LINE_COLOR.PURPLE,
            "Re-initializing boards after reset....",
            CMD_LINE_COLOR.ENDC,
        )
        try:
            if use_umd:
                options = TopologyDiscoveryOptions()
                options.no_wait_for_eth_training = not wait_for_eth
                if not wait_for_eth:
                    print(
                        CMD_LINE_COLOR.YELLOW,
                        "Skipping Ethernet link training wait after reset.",
                        CMD_LINE_COLOR.ENDC,
                    )
                TopologyDiscovery.discover(options)
            else:
                detect_chips_with_callback(print_status=print_status)
        except Exception as e:
            print(
                CMD_LINE_COLOR.RED,
                f"Error when re-initializing chips!\n {e}",
                CMD_LINE_COLOR.ENDC,
            )
            sys.exit(1)


def glx_6u_trays_reset(
    reinit: bool = True,
    ubb_num: str = "0xF",
    dev_num: str = "0xFF",
    op_mode: str = "0x0",
    reset_time: str = "0xF",
    print_status: bool = True,
    use_umd: bool = False,
):
    """
    Reset the WH asics on the galaxy systems with the following steps:
    1. Reset the trays with ipmi command (or UMD warm reset)
    2. Wait for 30s
    3. Reinit all chips

    Args:
        reinit: Whether to reinitialize the chips after reset.
        ubb_num: The UBB number to reset. 0x0~0xF (bit map)
        dev_num: The device number to reset. 0x0~0xFF(bit map)
        op_mode: The operation mode to use.
        reset_time: The reset time to use. resolution 10ms (ex. 0xF => 15 => 150ms)
        print_status: Whether to print out animations while detecting chips.
        use_umd: Whether to use UMD (WarmReset.ubb_warm_reset) or pyluwen (run_wh_ubb_ipmi_reset)
    """
    print(
        CMD_LINE_COLOR.PURPLE,
        "Resetting WH Galaxy trays with reset command...",
        CMD_LINE_COLOR.ENDC,
    )

    if use_umd:
        if ubb_num != "0xF" or dev_num != "0xFF" or op_mode != "0x0" or reset_time != "0xF":
            print(
                CMD_LINE_COLOR.RED,
                "Error: UMD warm reset only supports full galaxy reset (ubb_num=0xF, dev_num=0xFF, op_mode=0x0, reset_time=0xF)",
                CMD_LINE_COLOR.ENDC,
            )
            sys.exit(1)
        WarmReset.ubb_warm_reset(timeout_s=100.0)
        timed_wait(30)
        umd_ubb_wait_for_driver_load()
    else:
        run_wh_ubb_ipmi_reset(ubb_num, dev_num, op_mode, reset_time)
        timed_wait(30)
        run_ubb_wait_for_driver_load()

    print(
        CMD_LINE_COLOR.PURPLE,
        "Re-initializing boards after reset....",
        CMD_LINE_COLOR.ENDC,
    )
    if not reinit:
        print(
            CMD_LINE_COLOR.GREEN,
            "Exiting after galaxy reset without re-initializing chips.",
            CMD_LINE_COLOR.ENDC,
        )
        sys.exit(0)
    try:
        chips = detect_chips_with_callback(
            local_only=True, ignore_ethernet=True, print_status=print_status
        )
    except Exception as e:
        print(
            CMD_LINE_COLOR.RED,
            f"Error when re-initializing chips!\n {e}",
            CMD_LINE_COLOR.ENDC,
        )
        sys.exit(1)

    print(
        CMD_LINE_COLOR.GREEN,
        f"Re-initialized {len(chips)} boards after reset. Exiting...",
        CMD_LINE_COLOR.ENDC,
    )
    sys.exit(0)



# SPDX-FileCopyrightText: © 2026 Tenstorrent Inc.
# SPDX-License-Identifier: Apache-2.0

import pytest

from tt_smi.tt_smi_utils import get_board_type, convert_signed_16_16_to_float

class TestGetBoardType:
    @pytest.mark.parametrize(
        "upi,expected",
        [
            # Grayskull
            (0x3, "e150"),
            (0xA, "e300"),
            (0x7, "e75"),
            # Wormhole
            (0x8, "nb_cb"),
            (0xB, "wh_4u"),
            (0x14, "n300"),
            (0x18, "n150"),
            (0x35, "tt-galaxy-wh"),
            # Blackhole
            (0x36, "bh-scrappy"),
            (0x43, "p100a"),
            (0x40, "p150a"),
            (0x41, "p150b"),
            (0x42, "p150c"),
            (0x44, "p300b"),
            (0x45, "p300a"),
            (0x46, "p300c"),
            (0x47, "tt-galaxy-bh"),
        ],
    )
    def test_all_known_board_types(self, upi, expected):
        """Parametrized test for all known board types."""
        serial_num = upi << 36
        board_id = f"{serial_num:016x}"
        assert get_board_type(board_id) == expected

    @pytest.mark.requires_hardware
    def test_get_board_id_real_device(self, devices):
        """Test get_board_id with a real example."""
        for dev in devices.values():
            result = get_board_type(f"{dev.board_id():x}")
            # This is a real board ID, so just verify it returns a valid type
            assert result in [
                "e150",
                "e300",
                "e75",
                "nb_cb",
                "wh_4u",
                "n300",
                "n150",
                "tt-galaxy-wh",
                "bh-scrappy",
                "p100a",
                "p150a",
                "p150b",
                "p150c",
                "p300a",
                "p300b",
                "p300c",
                "tt-galaxy-bh",
            ]


class TestDataFormatting:
    @pytest.mark.parametrize(
        "raw,expected",
        [
            (0x00000000, 0.0),  # Zero
            (0x003C0000, 60.0),  # A typical ASIC temperature
            (0xFFF60000, -10.0),  # Negative value
            (0x7FFFFFFF, 32768.0),  # Max 16.16 positive value
            (0x80000000, -32768.0),  # Min 16.16 negative value
        ],
    )
    def test_convert_signed_16_16_to_float(self, raw, expected):
        """Test converting signed 16.16 fixed-point number to float."""
        assert convert_signed_16_16_to_float(raw) == pytest.approx(expected)



# SPDX-FileCopyrightText: © 2026 Tenstorrent Inc.
# SPDX-License-Identifier: Apache-2.0

import json
import pytest
import subprocess


@pytest.fixture(scope="session")
def snapshot(backend) -> dict:
    """Return snapshot data from TTSMIBackend."""
    log_str = backend.get_logs_json()
    log_json = json.loads(log_str)
    return log_json


@pytest.mark.requires_hardware
class TestSnapshot:
    """Test the tt-smi -s (snapshot) interface."""

    def test_snapshot_is_valid_json(self, snapshot):
        """Test if log string is valid json"""
        assert isinstance(snapshot, dict)

    def test_snapshot_fields_present(self, snapshot):
        """Test if fields are present in snapshot"""
        assert "time" in snapshot
        assert "host_info" in snapshot
        assert "host_sw_vers" in snapshot
        assert "device_info" in snapshot

    def test_host_info_fields_present(self, snapshot):
        """Test if fields are present in host_info"""
        host_info = snapshot["host_info"]
        assert "OS" in host_info
        assert "Distro" in host_info
        assert "Kernel" in host_info
        assert "Hostname" in host_info
        assert "Platform" in host_info
        assert "Python" in host_info
        assert "Memory" in host_info
        assert "Driver" in host_info

    def test_device_info_fields_present(self, snapshot):
        """Test if fields are present in device_info"""
        device_infos = snapshot["device_info"]
        for device_info in device_infos:
            assert "smbus_telem" in device_info
            assert "board_info" in device_info
            assert "telemetry" in device_info
            assert "firmwares" in device_info
            assert "limits" in device_info

    def test_smbus_telem_fields_present(self, snapshot):
        """Test if fields are present in smbus_telem."""
        smbus_telem_list = [
            "BOARD_ID_HIGH",
            "BOARD_ID_LOW",
            "ETH_FW_VERSION",
            "DDR_STATUS",
            "FAN_SPEED",
            # "FAN_RPM",
            "AICLK",
            "AXICLK",
            "ARCCLK",
            "VCORE",
            "ASIC_TEMPERATURE",
            "VREG_TEMPERATURE",
            "BOARD_TEMPERATURE",
            "TDP",
            "TDC",
            "VDD_LIMITS",
            "THM_LIMITS",
            "TT_FLASH_VERSION",
        ]
        device_infos = snapshot["device_info"]
        for device_info in device_infos:
            smbus_telem = device_info["smbus_telem"]
            for field in smbus_telem_list:
                assert field in smbus_telem

    def test_board_info_fields_present(self, snapshot):
        """Test if fields are present in board_info."""
        device_infos = snapshot["device_info"]
        for device_info in device_infos:
            board_info = device_info["board_info"]
            assert "bus_id" in board_info
            assert "board_type" in board_info
            assert "board_id" in board_info
            assert "coords" in board_info
            assert "dram_status" in board_info
            assert "dram_speed" in board_info
            assert "pcie_speed" in board_info
            assert "pcie_width" in board_info

    def test_telemetry_fields_present(self, snapshot):
        """Test if fields are present in telemetry."""
        device_infos = snapshot["device_info"]
        for device_info in device_infos:
            telemetry = device_info["telemetry"]
            assert "voltage" in telemetry
            assert "current" in telemetry
            assert "power" in telemetry
            assert "aiclk" in telemetry
            assert "asic_temperature" in telemetry
            assert "fan_speed" in telemetry
            assert "heartbeat" in telemetry

    def test_firmwares_fields_present(self, snapshot):
        """Test if fields are present in firmwares."""
        device_infos = snapshot["device_info"]
        for device_info in device_infos:
            firmwares = device_info["firmwares"]
            assert "fw_bundle_version" in firmwares
            assert "tt_flash_version" in firmwares
            assert "cm_fw" in firmwares
            assert "cm_fw_date" in firmwares
            assert "eth_fw" in firmwares
            assert "dm_bl_fw" in firmwares
            assert "dm_app_fw" in firmwares

    def test_limits_fields(self, snapshot):
        """Test if fields are present in limits."""
        device_infos = snapshot["device_info"]
        for device_info in device_infos:
            limits = device_info["limits"]
            assert "vdd_min" in limits
            assert "vdd_max" in limits
            assert "tdp_limit" in limits
            assert "tdc_limit" in limits
            assert "asic_fmax" in limits
            assert "therm_trip_l1_limit" in limits
            assert "thm_limit" in limits
            assert "bus_peak_limit" in limits

    def test_snapshot_no_tty(self):
        """Test if the output from tt-smi -s --snapshot_no_tty can be parsed as json"""
        result = subprocess.run(
            ["tt-smi", "-s", "--snapshot_no_tty"],
            capture_output=True,
            text=True,
        )
        log_json = json.loads(result.stdout)
        assert isinstance(log_json, dict)



# SPDX-FileCopyrightText: © 2026 Tenstorrent Inc.
# SPDX-License-Identifier: Apache-2.0

"""
Utility functions and constants for tt-smi (hex conversion, board type, logging path, etc.).
"""

import os
import sys
from importlib.metadata import version

from tt_tools_common.ui_common.themes import CMD_LINE_COLOR

from tt_smi import constants


LOG_FOLDER = os.path.expanduser("~/tt_smi_logs/")

def hex_to_date(hexdate: int, include_time=True):
    """Converts a date given in hex from format 0xYMDDHHMM to string YYYY-MM-DD HH:MM"""
    if hexdate == 0 or hexdate == 0xFFFFFFFF:
        return "N/A"

    year = (hexdate >> 28 & 0xF) + 2020
    month = hexdate >> 24 & 0xF
    day = hexdate >> 16 & 0xFF
    hour = hexdate >> 8 & 0xFF
    minute = hexdate & 0xFF

    date = f"{year:04}-{month:02}-{day:02}"

    if include_time:
        date += f" {hour:02}:{minute:02}"

    return date

def hex_to_semver_eth_wh(hexsemver: int):
    """
    Converts a semantic version string from format 0x061000 to 6.1.0
    Used in WH firmware only.
    """
    major = hexsemver >> 16 & 0xFF
    minor = hexsemver >> 12 & 0xF
    patch = hexsemver & 0xFFF

    return f"{major}.{minor}.{patch}"

def hex_to_semver_eth(hexsemver: int):
    """
    Converts a semantic version string from format 0x060100 to 6.1.0
    """
    major = hexsemver >> 16 & 0xFF
    minor = hexsemver >> 8 & 0xFF
    patch = hexsemver & 0xFF

    return f"{major}.{minor}.{patch}"

def hex_to_semver_m3_fw(hexsemver: int):
    """Converts a semantic version string from format 0x0A0F0100 to 10.15.1"""
    major = hexsemver >> 24 & 0xFF
    minor = hexsemver >> 16 & 0xFF
    patch = hexsemver >> 8 & 0xFF
    ver = hexsemver >> 0 & 0xFF

    return f"{major}.{minor}.{patch}.{ver}"


def get_board_type(board_id: str) -> str:
    """
    Get board type from board ID string.
    Ex:
        Board ID: AA-BBBBB-C-D-EE-FF-XXX
                   ^     ^ ^ ^  ^  ^   ^
                   |     | | |  |  |   +- XXX
                   |     | | |  |  +----- FF
                   |     | | |  +-------- EE
                   |     | | +----------- D
                   |     | +------------- C = Revision
                   |     +--------------- BBBBB = Unique Part Identifier (UPI)
                   +--------------------- AA
    """
    if board_id == "N/A":
        return "N/A"
    serial_num = int(f"0x{board_id}", base=16)
    upi = (serial_num >> 36) & 0xFFFFF

    # Grayskull cards
    if upi == 0x3:
        return "e150"
    elif upi == 0xA:
        return "e300"
    elif upi == 0x7:
        return "e75"

    # Wormhole cards
    elif upi == 0x8:
        return "nb_cb"
    elif upi == 0xB:
        return "wh_4u"
    elif upi == 0x14:
        return "n300"
    elif upi == 0x18:
        return "n150"
    elif upi == 0x35:
        return "tt-galaxy-wh"

    # Blackhole cards
    elif upi == 0x36:
        return "bh-scrappy"
    elif upi == 0x43:
        return "p100a"
    elif upi == 0x40:
        return "p150a"
    elif upi == 0x41:
        return "p150b"
    elif upi == 0x42:
        return "p150c"
    elif upi == 0x44:
        return "p300b"
    elif upi == 0x45:
        return "p300a"
    elif upi == 0x46:
        return "p300c"
    elif upi == 0x47:
        return "tt-galaxy-bh"
    else:
        return "N/A"


def convert_signed_16_16_to_float(value: int) -> float:
    """Convert signed 16.16 to float"""
    if value & (1 << (32 - 1)):  # if the value is negative (two's complement)
        value -= 1 << 32  # convert to negative value
    return value / 65536.0


def dict_from_public_attrs(obj: object) -> dict:
    """Parse an object's public attributes into a dictionary"""
    all_attrs = obj.__dir__()
    public = [attr for attr in all_attrs if not attr.startswith("_")]
    return {attr: getattr(obj, attr) for attr in public}


def get_host_software_versions() -> dict:
    """Return dict of tt_smi, pyluwen, tt_umd package versions."""
    return {
        "tt_smi": version("tt_smi"),
        "pyluwen": version("pyluwen"),
        "tt_umd": version("tt_umd"),
    }


def check_is_galaxy(backend, user_arg: str):
    """Check if the board is a Galaxy board."""
    if len(backend.device_infos) < 1:
        print(
            CMD_LINE_COLOR.RED,
            "No devices detected.",
            CMD_LINE_COLOR.ENDC,
        )
        sys.exit(1)

    if backend.device_infos[0]["board_type"] not in constants.GLX_BOARD_TYPES:
        print(
            CMD_LINE_COLOR.RED,
            f"This is not a Galaxy board, `{user_arg}` is only supported on Galaxy.",
            CMD_LINE_COLOR.ENDC,
        )
        sys.exit(1)


def is_vm() -> bool:
    """Return True if running inside a VM (e.g. hypervisor present in /proc/cpuinfo)."""
    try:
        with open("/proc/cpuinfo", "r") as f:
            if "hypervisor" in f.read():
                return True
    except (FileNotFoundError, PermissionError) as e:
        print(
            CMD_LINE_COLOR.YELLOW,
            f"Cannot access /proc/cpuinfo: {e}",
            CMD_LINE_COLOR.ENDC,
        )
    return False

def get_fw_bundle_version(smbus_telem_info) -> int:
    """We have two possible keys for the firmware bundle version: FLASH_BUNDLE_VERSION and FW_BUNDLE_VERSION.
    We need to return the version of the firmware bundle.
    """
    if "FW_BUNDLE_VERSION" in smbus_telem_info:
        return smbus_telem_info["FW_BUNDLE_VERSION"]
    elif "FLASH_BUNDLE_VERSION" in smbus_telem_info:
        return smbus_telem_info["FLASH_BUNDLE_VERSION"]
    else:
        return None



# SPDX-FileCopyrightText: © 2023 Tenstorrent Inc.
# SPDX-License-Identifier: Apache-2.0

"""
This is the backend of tt-smi.
    - Keeps track of chip objects and tasks related to fetching device info and telemetry
    - Sanitizes input for frontend
"""

import os
import re
import sys
import time
import datetime
from tt_smi import log
from pathlib import Path
from rich.table import Table
from tt_smi import constants
from rich import get_console
from rich.syntax import Syntax
from typing import Any, Dict, List, Optional, Union
from rich.progress import track
from tt_tools_common.ui_common.themes import CMD_LINE_COLOR
from pyluwen import PciChip
from tt_smi.tt_smi_utils import (
    LOG_FOLDER,
    hex_to_date,
    hex_to_semver_eth,
    hex_to_semver_m3_fw,
    hex_to_semver_eth_wh,
    get_board_type,
    convert_signed_16_16_to_float,
    dict_from_public_attrs,
    get_host_software_versions,
    get_fw_bundle_version
)
from tt_umd import (
    TTDevice,
    wormhole,
    TelemetryTag,
    ClusterDescriptor,
    SmBusArcTelemetryReader,
    ARCH,
)
from tt_tools_common.utils_common.system_utils import (
    get_host_info,
)
from tt_tools_common.utils_common.tools_utils import init_logging

class TTSMIBackend:
    """
    TT-SMI backend class that encompasses all chip objects on host.
    It handles all device related tasks like fetching device info, telemetry and toggling resets
    """

    def __init__(
        self,
        devices: Dict[int, Union[PciChip, TTDevice]] = None,
        umd_cluster_descriptor: Optional[ClusterDescriptor] = None,
        fully_init: bool = True,
        pretty_output: bool = True,
    ):
        self.devices = devices
        self.use_umd = umd_cluster_descriptor is not None
        if (self.use_umd):
            self.umd_cluster_descriptor = umd_cluster_descriptor
        self.pretty_output = pretty_output
        self.log: log.TTSMILog = log.TTSMILog(
            time=datetime.datetime.now(),
            host_info=get_host_info(),
            host_sw_vers=get_host_software_versions(),
            device_info=[
                log.TTSMIDeviceLog(
                    smbus_telem=log.SmbusTelem(),
                    board_info=log.BoardInfo(),
                    telemetry=log.Telemetry(),
                    firmwares=log.Firmwares(),
                    limits=log.Limits(),
                )
                for device in self.devices
            ],
        )
        self.smbus_telem_info = []
        self.firmware_infos = []
        self.device_infos = []
        self.device_telemetrys = []
        self.chip_limits = []
        self.pci_properties = []

        if fully_init:
            for i, _ in track(
                self.devices.items(),
                total=len(self.devices),
                description="Gathering Information",
                update_period=0.01,
                disable=not self.pretty_output,
            ):
                self.smbus_telem_info.append(self.get_smbus_board_info(i))
                self.firmware_infos.append(self.get_firmware_versions(i))
                self.pci_properties.append(self.get_pci_properties(i))
                self.device_infos.append(self.get_device_info(i))
                self.device_telemetrys.append(self.get_chip_telemetry(i))
                self.chip_limits.append(self.get_chip_limits(i))
    
    def is_blackhole(self, device_idx) -> bool:
        return (self.devices[device_idx].as_bh() if not self.use_umd
                else self.devices[device_idx].get_arch() == ARCH.BLACKHOLE)
    
    def is_wormhole(self, device_idx) -> bool:
        return (self.devices[device_idx].as_wh() if not self.use_umd
                else self.devices[device_idx].get_arch() == ARCH.WORMHOLE_B0)

    def is_grayskull(self, device_idx) -> bool:
        return (self.devices[device_idx].as_gs() if not self.use_umd
                else False)
    
    def get_pci_device_id(self, device_idx) -> str:
        if self.devices[device_idx].is_remote():
            return "N/A"
        return (self.devices[device_idx].get_pci_interface_id() if not self.use_umd
                else self.devices[device_idx].get_pci_device().get_device_num())
        
    def get_pci_bdf(self, device_idx) -> str:
        if self.devices[device_idx].is_remote():
            return "N/A"
        return (self.devices[device_idx].get_pci_bdf() if not self.use_umd
                else self.devices[device_idx].get_pci_device().get_device_info().pci_bdf)
    
    def get_device_name(self, device_idx):
        """Get device name from chip object"""
        if self.is_wormhole(device_idx):
            return "Wormhole"
        elif self.is_blackhole(device_idx):
            return "Blackhole"
        else:
            assert False, "Unknown chip name, FIX!"

    def save_logs_to_file(self, result_filename: str = ""):
        """Save log for smi snapshots"""
        time_now = datetime.datetime.now()
        date_string = time_now.strftime("%m-%d-%Y_%H:%M:%S")
        if not os.path.exists(LOG_FOLDER):
            init_logging(LOG_FOLDER)
        log_filename = f"{LOG_FOLDER}{date_string}_results.json"
        if result_filename:
            dir_path = os.path.dirname(os.path.realpath(result_filename))
            Path(dir_path).mkdir(parents=True, exist_ok=True)
            log_filename = result_filename

        clean_json = self.get_logs_json()
        with open(log_filename, "w") as f:
            f.write(clean_json)

        return log_filename

    def print_logs_to_stdout(self, pretty: bool = True):
        """Pretty-print (or just print) logs to stdout"""
        clean_json = self.get_logs_json()

        if not pretty:
            print(clean_json)
            return

        formatted = Syntax(clean_json, "json", background_color="default")
        console = get_console()
        console.print(formatted)

    def get_logs_json(self) -> str:
        """Get logs as JSON"""
        for i in self.devices:
            self.log.device_info[i].smbus_telem = self.smbus_telem_info[i]
            self.log.device_info[i].board_info = self.device_infos[i]
            # Add L/R for nb300 to separate local and remote asics
            if self.is_wormhole(i):
                board_type = self.device_infos[i]["board_type"]
                suffix = " R" if self.devices[i].is_remote() else " L"
                board_type = board_type + suffix
                self.log.device_info[i].board_info["board_type"] = board_type
            self.log.device_info[i].telemetry = self.device_telemetrys[i]
            self.log.device_info[i].firmwares = self.firmware_infos[i]
            self.log.device_info[i].limits = self.chip_limits[i]

        return self.log.get_clean_json_string()

    def print_all_available_devices_luwen(self):
        """Print all available boards on host (Luwen path). Shows both PCI BDF and PCI Dev ID. Doesn't show UMD Chip ID."""
        console = get_console()
        table_1 = Table(title="All available boards on host:")
        table_1.add_column("PCI BDF")
        table_1.add_column("PCI Dev ID")
        table_1.add_column("Board Type")
        table_1.add_column("Device Series")
        table_1.add_column("Board Number")
        for i in self.devices:
            board_id = self.device_infos[i]["board_id"]
            board_type = self.device_infos[i]["board_type"]
            pci_dev_id = self.get_pci_device_id(i)
            if self.is_wormhole(i):
                suffix = " R" if self.devices[i].is_remote() else " L"
                board_type = board_type + suffix

            table_1.add_row(
                f"{self.get_pci_bdf(i)}",
                f"/dev/tenstorrent/{i}",
                f"{self.get_device_name(i)}",
                f"{board_type}",
                f"{board_id}",
            )
        console.print(table_1)
        table_2 = Table(title="Boards that can be reset:")
        table_2.add_column("PCI BDF")
        table_2.add_column("PCI Dev ID")
        table_2.add_column("Board Type")
        table_2.add_column("Device Series")
        table_2.add_column("Board Number")
        for i in self.devices:
            if (
                not self.devices[i].is_remote()
                and self.device_infos[i]["board_type"] != "wh_4u"
            ):
                board_id = self.device_infos[i]["board_id"]
                board_type = self.device_infos[i]["board_type"]
                pci_dev_id = self.get_pci_device_id(i)
                if self.is_wormhole(i):
                    suffix = " R" if self.devices[i].is_remote() else " L"
                    board_type = board_type + suffix
                table_2.add_row(
                    f"{self.get_pci_bdf(i)}",
                    f"/dev/tenstorrent/{i}",
                    f"{self.get_device_name(i)}",
                    f"{board_type}",
                    f"{board_id}",
                )
        console.print(table_2)

    def print_all_available_devices_umd(self):
        """Print all available boards on host (UMD path). Shows UMD Chip ID, PCI BDF, and PCI Dev ID."""
        if not self.use_umd:
            raise RuntimeError("print_all_available_devices_umd requires UMD backend (umd_cluster_descriptor set)")
        console = get_console()
        table_1 = Table(title="All available boards on host (UMD):")
        table_1.add_column("UMD Chip ID")
        table_1.add_column("PCI BDF")
        table_1.add_column("PCI Dev ID")
        table_1.add_column("Board Type")
        table_1.add_column("Device Series")
        table_1.add_column("Board Number")
        for i in self.devices:
            pci_dev_num = self.get_pci_device_id(i)
            board_id = self.device_infos[i]["board_id"]
            board_type = self.device_infos[i]["board_type"]
            if self.is_wormhole(i):
                suffix = " R" if self.devices[i].is_remote() else " L"
                board_type = board_type + suffix
            table_1.add_row(
                f"{i}",
                f"{self.get_pci_bdf(i)}",
                f"/dev/tenstorrent/{pci_dev_num}",
                f"{self.get_device_name(i)}",
                f"{board_type}",
                f"{board_id}",
            )
        console.print(table_1)
        table_2 = Table(title="Boards that can be reset (UMD):")
        table_2.add_column("UMD Chip ID")
        table_2.add_column("PCI BDF")
        table_2.add_column("PCI Dev ID")
        table_2.add_column("Board Type")
        table_2.add_column("Device Series")
        table_2.add_column("Board Number")
        for i in self.devices:
            if (
                not self.devices[i].is_remote()
                and self.device_infos[i]["board_type"] != "wh_4u"
            ):
                pci_dev_num = self.get_pci_device_id(i)
                board_id = self.device_infos[i]["board_id"]
                board_type = self.device_infos[i]["board_type"]
                if self.is_wormhole(i):
                    suffix = " R" if self.devices[i].is_remote() else " L"
                    board_type = board_type + suffix
                table_2.add_row(
                    f"{i}",
                    f"{self.get_pci_bdf(i)}",
                    f"/dev/tenstorrent/{pci_dev_num}",
                    f"{self.get_device_name(i)}",
                    f"{board_type}",
                    f"{board_id}",
                )
        console.print(table_2)

    def print_tray_and_device_mapping(self):
        """Print the mapping of trays to devices on the galaxy"""

        ubb_bus_ids = (
            constants.BH_UBB_BUS_IDS
            if self.devices[0].as_bh()
            else constants.WH_UBB_BUS_IDS
        )

        def _get_mapping_tray_to_device():
            bus_id_to_tray = {
                bus_id: tray_num for tray_num, bus_id in ubb_bus_ids.items()
            }
            tray_to_devices = {}
            for i, device_info in enumerate(self.device_infos):
                # Extract bus id from "domain:bus:device.function" format
                bus_id = int(device_info["bus_id"].split(":")[1], 16)
                tray_bus_id = bus_id & 0xF0
                if tray_bus_id in bus_id_to_tray:
                    tray_num = bus_id_to_tray[tray_bus_id]
                    tray_to_devices.setdefault(tray_num, []).append(i)
            return tray_to_devices

        tray_to_devices = _get_mapping_tray_to_device()

        console = get_console()
        table = Table(title="Mapping of trays to devices on the galaxy:")
        table.add_column("Tray Number")
        table.add_column("Tray Bus ID")
        table.add_column("/dev/tenstorrent/<id>")
        for tray_num in sorted(tray_to_devices):
            table.add_row(
                f"{tray_num}",
                f"0x{ubb_bus_ids[tray_num]:02x}",
                f"{','.join(map(str, tray_to_devices[tray_num]))}",
            )
        console.print(table)

    def get_smbus_board_info(self, board_num: int) -> Dict:
        """Update board info by reading SMBUS_TELEMETRY"""
        if self.use_umd:
            smbus_telem_dict = {}
            arch = self.devices[board_num].get_arch()
            if arch == ARCH.WORMHOLE_B0:
                telem_reader = SmBusArcTelemetryReader(self.devices[board_num])
                tag_collection = wormhole.TelemetryTag
            elif arch == ARCH.BLACKHOLE:
                telem_reader = self.devices[board_num].get_arc_telemetry_reader()
                tag_collection = TelemetryTag
            else:
                raise ValueError(f"Unknown arch for device {board_num}")
            
            for telem_key in tag_collection:
                telem_value = hex(telem_reader.read_entry(telem_key.value)) if telem_reader.is_entry_available(telem_key.value) else None
                smbus_telem_dict[telem_key.name] = telem_value
            return smbus_telem_dict
        
        pyluwen_chip = self.devices[board_num]
        if pyluwen_chip.as_bh():
            telem_struct = pyluwen_chip.as_bh().get_telemetry()
        elif pyluwen_chip.as_wh():
            telem_struct = pyluwen_chip.as_wh().get_telemetry()
        else:
            raise ValueError(f"Unknown chip type for device {board_num}")
        json_map = dict_from_public_attrs(telem_struct)
        smbus_telem_dict = dict.fromkeys(constants.SMBUS_TELEMETRY_LIST)

        for key, value in json_map.items():
            if value is not None:
                value = hex(value)
            smbus_telem_dict[key.upper()] = value
        return smbus_telem_dict

    def update_telem(self):
        """Update telemetry in a given interval"""
        for i in self.devices:
            self.smbus_telem_info[i] = self.get_smbus_board_info(i)
            self.device_telemetrys[i] = self.get_chip_telemetry(i)

    def get_board_id(self, board_num) -> str:
        """Read board id from CSM or SPI if FW is not loaded"""
        if "BOARD_ID" in self.smbus_telem_info[board_num] and self.smbus_telem_info[board_num]["BOARD_ID"]:
            board_id = int(self.smbus_telem_info[board_num]["BOARD_ID"], base=16)
            return f"{board_id:016x}"
        else:
            board_info_0 = int(self.smbus_telem_info[board_num]["BOARD_ID_LOW"], base=16)
            board_info_1 = int(self.smbus_telem_info[board_num]["BOARD_ID_HIGH"], base=16)

            if board_info_0 is None or board_info_1 is None:
                return "N/A"
            return f"{board_info_1:08x}{board_info_0:08x}"

    def get_dram_speed(self, board_num) -> int:
        """Read DRAM Frequency from CSM and alternatively from SPI if FW not loaded on chip"""
        if self.is_blackhole(board_num):
            if self.smbus_telem_info[board_num]["DDR_SPEED"] is None:
                return "N/A"
            dram_speed = int(self.smbus_telem_info[board_num]["DDR_SPEED"], 16)
            # check if its div by 1000 and then return num GHz else MHz
            if dram_speed % 1000 == 0:
                dram_speed = dram_speed // 1000
                return f"{dram_speed}G"
            else:
                return f"{dram_speed}"
        elif self.is_wormhole(board_num):
            if self.smbus_telem_info[board_num]["DDR_STATUS"] is not None:
                dram_speed_raw = (
                    int(self.smbus_telem_info[board_num]["DDR_STATUS"], 16) >> 24
                )
                if dram_speed_raw == 0:
                    return "16G"
                elif dram_speed_raw == 1:
                    return "14G"
                elif dram_speed_raw == 2:
                    return "12G"
                elif dram_speed_raw == 3:
                    return "10G"
                elif dram_speed_raw == 4:
                    return "8G"
                else:
                    return None
            return "N/A"
        else:
            return "N/A"

    def get_pci_properties(self, board_num):
        """Get the PCI link speed and link width details from sysfs files"""
        if self.devices[board_num].is_remote():
            return {prop: "N/A" for prop in constants.PCI_PROPERTIES}

        try:
            pcie_bdf = self.get_pci_bdf(board_num)
            pci_bus_path = os.path.realpath(f"/sys/bus/pci/devices/{pcie_bdf}")
        except:
            return {prop: "N/A" for prop in constants.PCI_PROPERTIES}

        def get_pcie_gen(link_speed: str) -> int:
            if link_speed == "32.0":
                return 5
            if link_speed == "16.0":
                return 4
            elif link_speed == "8.0":
                return 3
            elif link_speed == "5.0":
                return 2
            elif link_speed == "2.5":
                return 1
            else:
                assert False, f"Invalid link speed {link_speed}"

        properties = {}

        for prop in constants.PCI_PROPERTIES:
            try:
                with open(os.path.join(pci_bus_path, prop), "r", encoding="utf-8") as f:
                    output = f.readline().rstrip()
                    value = re.findall(r"\d+\.\d+|\d+", output)[0]
                    if prop == "current_link_speed" or prop == "max_link_speed":
                        value = get_pcie_gen(value)
            except Exception:
                value = "N/A"
            properties[prop] = value
        return properties

    def get_dram_training_status(self, board_num) -> bool:
        """
        Get DRAM Training Status
        True means it passed training, False means it failed or did not train at all
        """
        if self.is_wormhole(board_num):
            # DDR_STATUS field in WH:
            # (31:24) dram speed
            # (23:0) per channel dram status (6 channels)
            # DRAM_TRAINING_FAIL = 0x1
            # DRAM_TRAINING_PASS = 0x2
            if self.smbus_telem_info[board_num]["DDR_STATUS"] is None:
                return False
            dram_status = (
                int(self.smbus_telem_info[board_num]["DDR_STATUS"], 16)) & 0xFFFFFF
            if dram_status == 0x222222:
                return True
            return False
        elif self.is_blackhole(board_num):
            dram_status = int(self.smbus_telem_info[board_num]["DDR_STATUS"], 16)
            if int(get_fw_bundle_version(self.smbus_telem_info[board_num]), 16) >= 0x13070000:
                # After FW version 19.7.0.3 (hex 0x13070003), the dram status is a 16-bit field with the following layout:
                # DDR Status:
                # [0]  - Training complete GDDR 0
                # [1]  - Error GDDR 0
                # [2]  - Training complete GDDR 1
                # [3]  - Error GDDR 1
                # ...
                # [14] - Training complete GDDR 7
                # [15] - Error GDDR 7
                # [16] - BIST complete GDDR 0
                # [17] - BIST failed GDDR 0
                # [18] - BIST complete GDDR 1
                # [19] - BIST failed GDDR 1
                # ...
                # [30] - BIST complete GDDR 7
                # [31] - BIST failed GDDR 7
                if dram_status == 0x55555555:
                    return True
                return False
            else:
                # Pre-19.7.0.3 DDR Status in BH is a 16-bit field with the following layout:
                #  [0] - Training complete GDDR 0
                #  [1] - Error GDDR 0
                #  [2] - Training complete GDDR 1
                #  [3] - Error GDDR 1
                #  ...
                #  [14] - Training Complete GDDR 7
                #  [15] - Error GDDR 7
                # 0x5555 = 0b0101010101010101 means all 8 channels trained successfully
                if dram_status == 0x5555:
                    return True
                return False
        else:
            return False

    def get_device_info(self, board_num) -> dict:
        dev_info = {}
        for field in constants.DEV_INFO_LIST:
            if field == "bus_id":
                try:
                    dev_info[field] = self.get_pci_bdf(board_num)
                except:
                    dev_info[field] = "N/A"
            elif field == "board_type":
                if self.get_board_id(board_num) == "N/A":
                    dev_info[field] = "N/A"
                else:
                    dev_info[field] = get_board_type(self.get_board_id(board_num))
            elif field == "board_id":
                dev_info[field] = self.get_board_id(board_num)
            elif field == "coords":
                if self.is_wormhole(board_num):
                    if self.use_umd:
                        if board_num in self.umd_cluster_descriptor.get_chip_locations():
                            eth_coord = self.umd_cluster_descriptor.get_chip_locations()[board_num]
                            dev_info[
                                field
                            ] = f"({eth_coord.x}, {eth_coord.y}, {eth_coord.rack}, {eth_coord.shelf})"
                        else:
                            dev_info[field] = "(0, 0, 0, 0)"
                    else:
                        dev_info[
                            field
                        ] = f"({self.devices[board_num].as_wh().get_local_coord().shelf_x}, {self.devices[board_num].as_wh().get_local_coord().shelf_y}, {self.devices[board_num].as_wh().get_local_coord().rack_x}, {self.devices[board_num].as_wh().get_local_coord().rack_y})"
                else:
                    dev_info[field] = "N/A"
            elif field == "dram_status":
                dev_info[field] = self.get_dram_training_status(board_num)
            elif field == "dram_speed":
                dev_info[field] = self.get_dram_speed(board_num)
            elif field == "pcie_speed":
                dev_info[field] = self.pci_properties[board_num]["current_link_speed"]
            elif field == "pcie_width":
                dev_info[field] = self.pci_properties[board_num]["current_link_width"]

        return dev_info

    def get_bh_chip_telemetry(self, board_num) -> Dict:
        """Get telemetry data for bh chip. None if ARC FW not running"""
        current = (
            int(self.smbus_telem_info[board_num]["TDC"], 16) & 0xFFFF
            if self.smbus_telem_info[board_num]["TDC"] is not None
            else 0
        )
        if self.smbus_telem_info[board_num]["VCORE"] is not None:
            voltage = int(self.smbus_telem_info[board_num]["VCORE"], 16) / 1000
        else:
            voltage = 10000
        power = (
            int(self.smbus_telem_info[board_num]["TDP"], 16) & 0xFFFF
            if self.smbus_telem_info[board_num]["TDP"] is not None
            else 0
        )
        asic_temperature = (
            (
                convert_signed_16_16_to_float(
                    int(self.smbus_telem_info[board_num]["ASIC_TEMPERATURE"], 16)
                )
            )
            if self.smbus_telem_info[board_num]["ASIC_TEMPERATURE"] is not None
            else 0
        )
        aiclk = (
            int(self.smbus_telem_info[board_num]["AICLK"], 16) & 0xFFFF
            if self.smbus_telem_info[board_num]["AICLK"] is not None
            else 0
        )
        timer_heartbeat = int(self.smbus_telem_info[board_num]["TIMER_HEARTBEAT"], 16) // 6 # Watchdog heartbeat, ~2 per second
        fan_speed = (
            min(int(self.smbus_telem_info[board_num]["FAN_RPM"], 16), 5000) / 50 # RPM to percent conversion
            if self.smbus_telem_info[board_num]["FAN_RPM"] is not None
            else 0
        )

        chip_telemetry = {
            "voltage": f"{voltage:4.2f}",
            "current": f"{current:5.1f}",
            "power": f"{power:5.1f}",
            "aiclk": f"{aiclk:4.0f}",
            "asic_temperature": f"{asic_temperature:4.1f}",
            "fan_speed": f"{fan_speed:3.0f}",
            "heartbeat": f"{timer_heartbeat}",
        }
        return chip_telemetry

    def get_wh_chip_telemetry(self, board_num) -> Dict:
        """Get telemetry data for WH chip. None if ARC FW not running"""
        current = int(self.smbus_telem_info[board_num]["TDC"], 16) & 0xFFFF
        if self.smbus_telem_info[board_num]["VCORE"] is not None:
            voltage = int(self.smbus_telem_info[board_num]["VCORE"], 16) / 1000
        else:
            voltage = 10000
        power = int(self.smbus_telem_info[board_num]["TDP"], 16) & 0xFFFF
        asic_temperature = (
            int(self.smbus_telem_info[board_num]["ASIC_TEMPERATURE"], 16) & 0xFFFF
        ) / 16
        aiclk = int(self.smbus_telem_info[board_num]["AICLK"], 16) & 0xFFFF
        arc3_heartbeat = int(self.smbus_telem_info[board_num]["ARC3_HEALTH"], 16) // 5 # Watchdog heartbeat, ~2 per second
        if self.smbus_telem_info[board_num]["FAN_SPEED"] is not None:
            fan_speed = int(self.smbus_telem_info[board_num]["FAN_SPEED"], 16)
        else:
            fan_speed = 0

        chip_telemetry = {
            "voltage": f"{voltage:4.2f}",
            "current": f"{current:5.1f}",
            "power": f"{power:5.1f}",
            "aiclk": f"{aiclk:4.0f}",
            "asic_temperature": f"{asic_temperature:4.1f}",
            "fan_speed": f"{fan_speed:3.0f}",
            "heartbeat": f"{arc3_heartbeat}"
        }

        return chip_telemetry

    def get_chip_telemetry(self, board_num) -> Dict:
        """Return the correct chip telemetry for a given board"""
        if self.is_blackhole(board_num):
            return self.get_bh_chip_telemetry(board_num)
        elif self.is_wormhole(board_num):
            return self.get_wh_chip_telemetry(board_num)
        else:
            print(
                CMD_LINE_COLOR.RED,
                f"Could not fetch telemetry for board {e}: Unrecognized board type!",
                CMD_LINE_COLOR.ENDC,
            )
            return {}

    def get_chip_limits(self, board_num: int) -> Dict[str, str]:
        if self.is_blackhole(board_num):
            return self.get_bh_chip_limits(board_num)
        elif self.is_wormhole(board_num):
            return self.get_wh_chip_limits(board_num)
        else:
            print(
                CMD_LINE_COLOR.RED,
                f"Could not fetch chip limits for board {board_num}: Unrecognized board type!",
                CMD_LINE_COLOR.ENDC,
            )
            return {}

    def get_wh_chip_limits(self, board_num: int) -> Dict[str, str]:
        """Get chip limits from the CSM. None if ARC FW not running"""

        chip_limits = {}
        for field in constants.LIMITS:
            if field == "vdd_min":
                vdd_limits = self.smbus_telem_info[board_num].get("VDD_LIMITS")
                chip_limits[field] = f"{(int(vdd_limits, 16) & 0xFFFF) / 1000:4.2f}" if vdd_limits else 0
            elif field == "vdd_max":
                vdd_limits = self.smbus_telem_info[board_num].get("VDD_LIMITS")
                chip_limits[field] = f"{(int(vdd_limits, 16) >> 16) / 1000:4.2f}" if vdd_limits else 0
            elif field == "tdp_limit":
                tdp = self.smbus_telem_info[board_num].get("TDP")
                chip_limits[field] = f"{int(tdp, 16) >> 16:3.0f}" if tdp else 0
            elif field == "tdc_limit":
                tdc = self.smbus_telem_info[board_num].get("TDC")
                chip_limits[field] = f"{int(tdc, 16) >> 16:3.0f}" if tdc else 0
            elif field == "asic_fmax":
                aiclk = self.smbus_telem_info[board_num].get("AICLK")
                chip_limits[field] = f"{int(aiclk, 16) >> 16:4.0f}" if aiclk else 0
            elif field == "therm_trip_l1_limit":
                thm_limits = self.smbus_telem_info[board_num].get("THM_LIMITS")
                chip_limits[field] = f"{int(thm_limits, 16) >> 16:2.0f}" if thm_limits else 0
            elif field == "thm_limit":
                thm_limits = self.smbus_telem_info[board_num].get("THM_LIMITS")
                chip_limits[field] = f"{int(thm_limits, 16) & 0xFFFF:2.0f}" if thm_limits else 0
            else:
                chip_limits[field] = 0
        return chip_limits

    def get_bh_chip_limits(self, board_num: int) -> Dict[str, str]:
        """Get chip limits from BH telemetry. None if ARC FW not running"""

        chip_limits = {}
        for field in constants.LIMITS:
            if field == "vdd_min":
                vdd_limits = self.smbus_telem_info[board_num].get("VDD_LIMITS")
                chip_limits[field] = f"{(int(vdd_limits, 16) & 0xFFFF) / 1000:4.2f}" if vdd_limits else 0
            elif field == "vdd_max":
                vdd_limits = self.smbus_telem_info[board_num].get("VDD_LIMITS")
                chip_limits[field] = f"{(int(vdd_limits, 16) >> 16) / 1000:4.2f}" if vdd_limits else 0
            elif field == "tdp_limit":
                tdp_limit = self.smbus_telem_info[board_num].get("TDP_LIMIT_MAX")
                chip_limits[field] = f"{int(tdp_limit, 16):3.0f}" if tdp_limit else 0
            elif field == "tdc_limit":
                tdc_limit = self.smbus_telem_info[board_num].get("TDC_LIMIT_MAX")
                chip_limits[field] = f"{int(tdc_limit, 16):3.0f}" if tdc_limit else 0
            elif field == "asic_fmax":
                asic_fmax = self.smbus_telem_info[board_num].get("AICLK_LIMIT_MAX")
                chip_limits[field] = f"{int(asic_fmax, 16):4.0f}" if asic_fmax else 0
            elif field == "therm_trip_l1_limit":
                therm_trip_l1_limit = self.smbus_telem_info[board_num].get("THM_LIMIT_THROTTLE")
                chip_limits[field] = f"{int(therm_trip_l1_limit, 16):2.0f}" if therm_trip_l1_limit else 0
            elif field == "thm_limit":
                thm_limits = self.smbus_telem_info[board_num].get("THM_LIMITS")
                chip_limits[field] = f"{int(thm_limits, 16):2.0f}" if thm_limits else 0
            else:
                chip_limits[field] = 0
        return chip_limits

    def get_firmware_versions(self, board_num):
        """Translate the telem struct semver for gui"""
        fw_versions = {}
        for field in constants.FW_LIST_SNAPSHOT:
            if field == "cm_fw":
                if "ARC0_FW_VERSION" in self.smbus_telem_info[board_num]:
                    val = self.smbus_telem_info[board_num]["ARC0_FW_VERSION"]
                elif "CM_FW_VERSION" in self.smbus_telem_info[board_num]:
                    val = self.smbus_telem_info[board_num]["CM_FW_VERSION"]
                else:
                    val = "N/A"
                if val is None:
                    fw_versions[field] = "N/A"
                else:
                    fw_versions[field] = hex_to_semver_m3_fw(int(val, 16))

            elif field == "cm_fw_date":
                if "WH_FW_DATE" in self.smbus_telem_info[board_num]:
                    val = self.smbus_telem_info[board_num]["WH_FW_DATE"]
                if val is None:
                    fw_versions[field] = "N/A"
                else:
                    fw_versions[field] = hex_to_date(int(val, 16), include_time=False)

            elif field == "eth_fw":
                val = self.smbus_telem_info[board_num]["ETH_FW_VERSION"]
                if val is None:
                    fw_versions[field] = "N/A"
                elif self.is_wormhole(board_num):
                    fw_versions[field] = hex_to_semver_eth_wh(int(val, 16))
                else:
                    fw_versions[field] = hex_to_semver_eth(int(val, 16))
            elif field == "dm_bl_fw":
                if self.use_umd:
                    # The tag has different value for old WH telemetry and new telemetry.
                    if "M3_BL_FW_VERSION" in self.smbus_telem_info[board_num]:
                        val = self.smbus_telem_info[board_num]["M3_BL_FW_VERSION"]
                    if "DM_BL_FW_VERSION" in self.smbus_telem_info[board_num]:
                        val = self.smbus_telem_info[board_num]["DM_BL_FW_VERSION"]
                else:
                    val = self.smbus_telem_info[board_num]["M3_BL_FW_VERSION"]
                if val is None:
                    fw_versions[field] = "N/A"
                else:
                    fw_versions[field] = hex_to_semver_m3_fw(int(val, 16))
            elif field == "dm_app_fw":
                if self.use_umd:
                    # The tag has different value for WH and BH
                    if "M3_APP_FW_VERSION" in self.smbus_telem_info[board_num]:
                        val = self.smbus_telem_info[board_num]["M3_APP_FW_VERSION"]
                    if "DM_APP_FW_VERSION" in self.smbus_telem_info[board_num]:
                        val = self.smbus_telem_info[board_num]["DM_APP_FW_VERSION"]
                else:
                    val = self.smbus_telem_info[board_num]["M3_APP_FW_VERSION"]
                if val is None:
                    fw_versions[field] = "N/A"
                else:
                    fw_versions[field] = hex_to_semver_m3_fw(int(val, 16))
            elif field == "tt_flash_version":
                if "TT_FLASH_VERSION" in self.smbus_telem_info[board_num]:
                    val = self.smbus_telem_info[board_num]["TT_FLASH_VERSION"]
                if val is None:
                    fw_versions[field] = "N/A"
                # See below- Galaxy systems manually get an N/A tt_flash_version
                elif get_board_type(self.get_board_id(board_num)) == "wh_4u":
                    fw_versions[field] = "N/A"
                else:
                    fw_versions[field] = hex_to_semver_m3_fw(int(val, 16))
            elif field == "fw_bundle_version":
                val = get_fw_bundle_version(self.smbus_telem_info[board_num])
                if (
                    get_board_type(self.get_board_id(board_num)) == "wh_4u"
                    and val == "0xffffffff"
                ):
                    # WARNING: Dirty dirty hack!
                    # See Issue #72 https://github.com/tenstorrent/tt-smi/issues/72
                    # Due to a FW flashing bug, Galaxy systems do not have the FW_BUNDLE_VERSION
                    # field set. The value ends up in TT_FLASH_VERSION, so we need to go get it.
                    val = self.smbus_telem_info[board_num]["TT_FLASH_VERSION"]
                    fw_versions[field] = hex_to_semver_m3_fw(int(val, 16))
                if val is None:
                    fw_versions[field] = "N/A"
                else:
                    fw_versions[field] = hex_to_semver_m3_fw(int(val, 16))
        return fw_versions



# SPDX-FileCopyrightText: © 2025 Tenstorrent Inc.
# SPDX-License-Identifier: Apache-2.0

import pytest

from typing import List

from pyluwen import pci_scan
from tt_umd import PCIDevice
from tt_smi.tt_smi_reset import pci_board_reset, glx_6u_trays_reset

NUM_RESETS_STRESS_TEST = 10


def redetect_devices(use_umd: bool) -> List[int]:
    if use_umd:
        return list(PCIDevice.enumerate_devices_info())
    return pci_scan()


@pytest.mark.requires_hardware
class TestPciDriverReset:
    def test_pci_reset_all_devices(self, reset_test_config):
        """
        Test resetting all PCI devices (invoked by tt-smi -r).

        Passes if the reset is successful and the same number of devices are
        detected before and after.
        """
        pci_indices, use_umd = reset_test_config
        pci_board_reset(pci_indices, reinit=True, use_umd=use_umd)
        post_reset_devices = redetect_devices(use_umd)
        assert len(post_reset_devices) == len(pci_indices)

    def test_pci_reset_all_devices_stress(self, reset_test_config):
        """
        Test resetting all PCI devices NUM_RESETS_STRESS_TEST times in a row.

        Passes if the resets are successful and the same number of devices are
        detected before and after each reset.
        """
        pci_indices, use_umd = reset_test_config
        for _ in range(NUM_RESETS_STRESS_TEST):
            pci_board_reset(pci_indices, reinit=True, use_umd=use_umd)
            post_reset_devices = redetect_devices(use_umd)
            assert len(post_reset_devices) == len(pci_indices)


@pytest.mark.requires_hardware
@pytest.mark.requires_galaxy
class TestGalaxyReset:
    def test_glx_reset(self, reset_test_config):
        """
        Test galaxy 6U trays reset (invoked by tt-smi -glx_reset).

        Passes if the reset is successful and the same number of devices are
        detected before and after.
        """
        pci_indices, use_umd = reset_test_config
        # Expect SystemExit with return code 0 on successful reset
        with pytest.raises(SystemExit) as exc_info:
            glx_6u_trays_reset(use_umd=use_umd)

        assert exc_info.value.code == 0

        post_reset_devices = redetect_devices(use_umd)
        assert len(post_reset_devices) == len(pci_indices)

    def test_glx_reset_stress(self, reset_test_config):
        """
        Test galaxy 6U trays reset NUM_RESETS_STRESS_TEST times in a row.

        Passes if the resets are successful and the same number of devices are
        detected before and after each reset.
        """
        pci_indices, use_umd = reset_test_config
        for _ in range(NUM_RESETS_STRESS_TEST):
            # Expect SystemExit with return code 0 on successful reset
            with pytest.raises(SystemExit) as exc_info:
                glx_6u_trays_reset(use_umd=use_umd)

            assert exc_info.value.code == 0

            post_reset_devices = redetect_devices(use_umd)
            assert len(post_reset_devices) == len(pci_indices)
